==> MRAConnector-0.91.tar.bz2: MRAConnector-0.91/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapOpenedMailboxInfo.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapOpenedMailboxInfo : ImapMailboxInfo, IDisposable {
    public override bool IsOpen {
      get { return Client.OpenedMailbox == this; }
    }

    public bool IsReadOnly {
      get { return Mailbox.ReadOnly; }
    }

    public long FirstUnseenMessageNumber {
      get { return Mailbox.FirstUnseen; }
    }

    public IImapMessageFlagSet ApplicableFlags {
      get { return Mailbox.ApplicableFlags; }
    }

    public IImapMessageFlagSet PermanentFlags {
      get { return Mailbox.PermanentFlags; }
    }

    public bool IsAllowedToCreateKeywords {
      get { return Mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords); }
    }

    public bool IsUidPersistent {
      get { return Mailbox.UidPersistent; }
    }

    private ImapMessageInfoCollection messages = new ImapMessageInfoCollection();

    internal ImapOpenedMailboxInfo(ImapClient client, ImapMailbox mailbox)
      : base(client, mailbox)
    {
    }

    void IDisposable.Dispose()
    {
      if (Client.IsConnected && IsOpen)
        Client.CloseMailbox();
    }

    public override ImapOpenedMailboxInfo Open(bool asReadOnly)
    {
      if (IsOpen)
        return this;
      else
        // messages.Clear();
        throw new NotImplementedException();
    }

    public void Close()
    {
      if (IsOpen)
        Client.CloseMailbox();
    }

    internal void CheckSelected()
    {
      if (!IsOpen)
        throw new ImapUnavailableException();
    }

    internal void CheckUidValidity(long uidValidity, ImapSequenceSet sequenceOrUidSet)
    {
      // TODO: impl
#if false
      if (!sequenceOrUidSet.IsUidSet)
        return;
      if (uidValidity != UidValidity)
        throw new ImapException("UIDVALIDITY value has been changed");
#endif
    }

    internal ImapMessageInfo ToMessageInfo(ImapMessageAttribute message, bool hasStaticAttr, bool hasDynamicAttr)
    {
      var info = messages.FindByUid(message.Uid);

      if (info == null) {
        info = new ImapMessageInfo(this, message.Uid, message.Sequence);

        messages.Add(info);
      }
      else if (info.Sequence != message.Sequence) {
        // XXX: this would be never occur?
        messages.ChangeSequenceNumber(info, message.Sequence);
      }

      if (hasStaticAttr)
        info.StaticAttribute = message.GetStaticAttributeImpl();
      if (hasDynamicAttr)
        info.DynamicAttribute = message.GetDynamicAttributeImpl();

      return info;
    }

    internal void Detach(ImapMessageInfoBase mailboxObject)
    {
      mailboxObject.DetachFromMailbox();

      var message = mailboxObject as ImapMessageInfo;

      if (message != null)
        messages.Remove(message);
    }

    internal void DetachAllMailboxObject()
    {
      foreach (var message in messages) {
        message.DetachFromMailbox();
      }

      messages.Clear();
    }

    internal ImapCommandResult ProcessMessageStatusChangedResult(ImapCommandResult result,
                                                                 ImapMessageAttributeBase[] statusChangedMessageAttrs)
    {
      ProcessSizeAndStatusResponse(result.ReceivedResponses, false);

      if (statusChangedMessageAttrs != null) {
        RaiseMessageStatusChanged(statusChangedMessageAttrs.Select(delegate(ImapMessageAttributeBase attr) {
          return attr.Sequence;
        }));
      }

      return base.ProcessResult(result, null);
    }

    internal protected override ImapCommandResult ProcessResult(ImapCommandResult result,
                                                                Func<ImapResponseCode, bool> throwIfError)
    {
      ProcessSizeAndStatusResponse(result.ReceivedResponses, true);

      return base.ProcessResult(result, throwIfError);
    }

    private void ProcessSizeAndStatusResponse(IEnumerable<ImapResponse> receivedResponses, bool processFetchResponse)
    {
      var prevExistMessageCount = Mailbox.ExistsMessage;
      var prevRecentMessageCount = Mailbox.RecentMessage;
      List<ImapMessageInfo> deletedMessages = null;
      List<long> statusChangedSequenceNumbers = null;

      foreach (var response in receivedResponses) {
        var data = response as ImapDataResponse;

        if (data == null)
          continue;

        // XXX: equality
        if (processFetchResponse && data.Type == ImapDataResponseType.Fetch) {
          if (statusChangedSequenceNumbers == null)
            statusChangedSequenceNumbers = new List<long>();

          statusChangedSequenceNumbers.Add(ImapDataResponseConverter.FromFetch(messages, data));
        }
        else if (data.Type == ImapDataResponseType.Expunge) {
          var expunged = ImapDataResponseConverter.FromExpunge(data);

          foreach (var message in messages) {
            if (message.Sequence == expunged) {
              if (deletedMessages == null)
                deletedMessages = new List<ImapMessageInfo>();

              deletedMessages.Add(message);
            }
            else if (expunged < message.Sequence) {
              message.Sequence--;
            }
          }

          if (0L < Mailbox.ExistsMessage)
            Mailbox.ExistsMessage -= 1L;
        }
        else if (data.Type == ImapDataResponseType.Exists) {
          Mailbox.ExistsMessage = ImapDataResponseConverter.FromExists(data);
        }
        else if (data.Type == ImapDataResponseType.Recent) {
          Mailbox.RecentMessage = ImapDataResponseConverter.FromRecent(data);
        }
        else if (data.Type == ImapDataResponseType.Flags) {
          Mailbox.ApplicableFlags = ImapDataResponseConverter.FromFlags(data);
        }
      }

      if (deletedMessages != null) {
        foreach (var message in deletedMessages) {
          message.SetDeleted();
        }
      }

      if (prevExistMessageCount != Mailbox.ExistsMessage)
        Client.RaiseExistMessageCountChanged(this, prevExistMessageCount);

      if (prevRecentMessageCount != Mailbox.RecentMessage)
        Client.RaiseRecentMessageCountChanged(this, prevRecentMessageCount);

      if (statusChangedSequenceNumbers != null)
        RaiseMessageStatusChanged(statusChangedSequenceNumbers);

      if (deletedMessages != null)
        Client.RaiseMessageDeleted(deletedMessages.ToArray());
    }

    private void RaiseMessageStatusChanged(IEnumerable<long> statusChangedSequenceNumbers)
    {
      var statusChangedMessages = new List<ImapMessageInfo>();

      foreach (var sequenceNumber in statusChangedSequenceNumbers) {
        var message = messages.FindBySequence(sequenceNumber);

        if (message != null)
          statusChangedMessages.Add(message);
      }

      Client.RaiseMessageStatusChanged(statusChangedMessages.ToArray());
    }

    /*
     * operations
     */
    public override void Refresh()
    {
      CheckSelected();

      ProcessResult(Client.Session.NoOp());
    }

    public void Expunge()
    {
      CheckSelected();

      ProcessResult(Client.Session.Expunge());
    }

    /*
     * MoveMessagesTo
     */
    public void MoveMessagesTo(ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapMailboxInfo destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(ImapMessageFetchAttributeOptions.None,
                                           uid,
                                           uids),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(searchCriteria,
                                           null,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(searchCriteria,
                                           encoding,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    /*
     * CopyMessagesTo
     */
    public void CopyMessagesTo(ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapMailboxInfo destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(ImapMessageFetchAttributeOptions.None,
                                           uid,
                                           uids),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(searchCriteria,
                                           null,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(searchCriteria,
                                           encoding,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    private void MoveOrCopyMessagesToCore(bool copy,
                                          ImapMessageInfoList messages,
                                          ImapMailboxInfo destinationMailbox)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      if (copy)
        messages.CopyTo(destinationMailbox);
      else
        messages.MoveTo(destinationMailbox);
    }
  }
}

==> MRAConnector-1.12.tar.bz2: MRAConnector-1.12/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client/ImapOpenedMailboxInfo.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapOpenedMailboxInfo : ImapMailboxInfo, IDisposable {
    public override bool IsOpen {
      get { return Client.OpenedMailbox == this; }
    }

    public bool IsReadOnly {
      get { return Mailbox.ReadOnly; }
    }

    public long FirstUnseenMessageNumber {
      get { return Mailbox.FirstUnseen; }
    }

    public IImapMessageFlagSet ApplicableFlags {
      get { return Mailbox.ApplicableFlags; }
    }

    public IImapMessageFlagSet PermanentFlags {
      get { return Mailbox.PermanentFlags; }
    }

    public bool IsAllowedToCreateKeywords {
      get { return Mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords); }
    }

    public bool IsUidPersistent {
      get { return Mailbox.UidPersistent; }
    }

    private ImapMessageInfoCollection messages = new ImapMessageInfoCollection();

    internal ImapOpenedMailboxInfo(ImapClient client, ImapMailbox mailbox)
      : base(client, mailbox)
    {
    }

    void IDisposable.Dispose()
    {
      if (Client.IsConnected && IsOpen)
        Client.CloseMailbox();
    }

    public override ImapOpenedMailboxInfo Open(bool asReadOnly)
    {
      if (IsOpen)
        return this;
      else
        // messages.Clear();
        throw new NotImplementedException();
    }

#if ASYNC_OPS
    public override Task<ImapOpenedMailboxInfo> OpenAsync(bool asReadOnly)
    {
      if (IsOpen) {
        var tcs = new TaskCompletionSource<ImapOpenedMailboxInfo>();

        tcs.SetResult(this);

        return tcs.Task;
      }
      else {
        // messages.Clear();
        throw new NotImplementedException();
      }
    }
#endif

    public void Close()
    {
      if (IsOpen)
        Client.CloseMailbox();
    }

#if ASYNC_OPS
    public Task CloseAsync()
    {
      if (IsOpen) {
        return Client.CloseMailboxAsync();
      }
      else {
        // do nothing
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }
    }
#endif

    internal void CheckSelected()
    {
      if (!IsOpen)
        throw new ImapUnavailableException();
    }

    internal void CheckUidValidity(long uidValidity, ImapSequenceSet sequenceOrUidSet)
    {
      // TODO: impl
#if false
      if (!sequenceOrUidSet.IsUidSet)
        return;
      if (uidValidity != UidValidity)
        throw new ImapException("UIDVALIDITY value has been changed");
#endif
    }

    internal ImapMessageInfo ToMessageInfo(ImapMessageAttribute message, bool hasStaticAttr, bool hasDynamicAttr)
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      var info = messages.FindByUid(message.Uid);

      if (info == null) {
        info = new ImapMessageInfo(this, message.Uid, message.Sequence);

        messages.Add(info);
      }
      else if (info.Sequence != message.Sequence) {
        // XXX: this would be never occur?
        messages.ChangeSequenceNumber(info, message.Sequence);
      }

      if (hasStaticAttr)
        info.StaticAttribute = message.GetStaticAttributeImpl();
      if (hasDynamicAttr)
        info.DynamicAttribute = message.GetDynamicAttributeImpl();

      return info;
    }

    internal void Detach(ImapMessageInfoBase mailboxObject)
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      mailboxObject.DetachFromMailbox();

      var message = mailboxObject as ImapMessageInfo;

      if (message != null)
        messages.Remove(message);
    }

    internal void DetachAllMailboxObject()
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      foreach (var message in messages) {
        message.DetachFromMailbox();
      }

      messages.Clear();
    }

    internal ImapCommandResult ProcessMessageStatusChangedResult(ImapCommandResult result,
                                                                 ImapMessageAttributeBase[] statusChangedMessageAttrs)
    {
      ProcessSizeAndStatusResponse(result.ReceivedResponses, false);

      if (statusChangedMessageAttrs != null) {
        QueueMessageStatusChangedEvent(statusChangedMessageAttrs.Select(delegate(ImapMessageAttributeBase attr) {
          return attr.Sequence;
        }));
      }

      return base.ProcessResult(result, null);
    }

    internal protected override ImapCommandResult ProcessResult(ImapCommandResult result,
                                                                Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse)
    {
      ProcessSizeAndStatusResponse(result.ReceivedResponses, true);

      return base.ProcessResult(result, throwIfErrorResponse);
    }

    private void ProcessSizeAndStatusResponse(IEnumerable<ImapResponse> receivedResponses, bool processFetchResponse)
    {
      var prevExistMessageCount = Mailbox.ExistsMessage;
      var prevRecentMessageCount = Mailbox.RecentMessage;
      List<ImapMessageInfo> deletedMessages = null;
      List<long> statusChangedSequenceNumbers = null;

      foreach (var response in receivedResponses) {
        var data = response as ImapDataResponse;

        if (data == null)
          continue;

        if (processFetchResponse && data.Type == ImapDataResponseType.Fetch) {
          var statusChangedSequenceNumber = ImapDataResponseConverter.FromFetch(messages, data);

          if (statusChangedSequenceNumber == 0L)
            continue;

          if (statusChangedSequenceNumbers == null)
            statusChangedSequenceNumbers = new List<long>();

          statusChangedSequenceNumbers.Add(statusChangedSequenceNumber);
        }
        else if (data.Type == ImapDataResponseType.Expunge) {
          var expunged = ImapDataResponseConverter.FromExpunge(data);

          foreach (var message in messages) {
            if (message.Sequence == expunged) {
              if (deletedMessages == null)
                deletedMessages = new List<ImapMessageInfo>();

              deletedMessages.Add(message);
            }
            else if (expunged < message.Sequence) {
              message.Sequence--;
            }
          }

          if (0L < Mailbox.ExistsMessage)
            Mailbox.ExistsMessage -= 1L;
        }
        else if (data.Type == ImapDataResponseType.Exists) {
          Mailbox.ExistsMessage = ImapDataResponseConverter.FromExists(data);
        }
        else if (data.Type == ImapDataResponseType.Recent) {
          Mailbox.RecentMessage = ImapDataResponseConverter.FromRecent(data);
        }
        else if (data.Type == ImapDataResponseType.Flags) {
          Mailbox.ApplicableFlags = ImapDataResponseConverter.FromFlags(data);
        }
      }

      if (deletedMessages != null) {
        foreach (var message in deletedMessages) {
          message.SetDeleted();
        }
      }

      if (prevExistMessageCount != Mailbox.ExistsMessage)
        Client.RaiseEvent(ImapClient.EventType.ExistMessageCountChanged,
                          new ImapMailboxSizeChangedEventArgs(this, Mailbox.ExistsMessage, prevExistMessageCount),
                          IsIdling);

      if (prevRecentMessageCount != Mailbox.RecentMessage)
        Client.RaiseEvent(ImapClient.EventType.RecentMessageCountChanged,
                          new ImapMailboxSizeChangedEventArgs(this, Mailbox.RecentMessage, prevRecentMessageCount),
                          IsIdling);

      if (statusChangedSequenceNumbers != null)
        QueueMessageStatusChangedEvent(statusChangedSequenceNumbers);

      if (deletedMessages != null)
        Client.RaiseEvent(ImapClient.EventType.MessageDeleted,
                          new ImapMessageStatusChangedEventArgs(deletedMessages.ToArray()),
                          IsIdling);
    }

    private void QueueMessageStatusChangedEvent(IEnumerable<long> statusChangedSequenceNumbers)
    {
      var statusChangedMessages = new List<ImapMessageInfo>();

      foreach (var sequenceNumber in statusChangedSequenceNumbers) {
        var message = messages.FindBySequence(sequenceNumber);

        if (message != null)
          statusChangedMessages.Add(message);
      }

      if (statusChangedMessages.Count == 0)
        return;

      Client.RaiseEvent(ImapClient.EventType.MessageStatusChanged,
                        new ImapMessageStatusChangedEventArgs(statusChangedMessages.ToArray()),
                        IsIdling);
    }

    /*
     * operations
     */
    internal protected override void RefreshCore(bool reretrieveMailboxFlags)
    {
      CheckSelected();

      if (reretrieveMailboxFlags)
        ReRetrieveMailboxFlags();

      ProcessResult(Client.Session.NoOp());
    }

    public void Expunge()
    {
      CheckSelected();

      Client.RunExclusively(() => ProcessResult(Client.Session.Expunge()));
    }

#if ASYNC_OPS
    public Task ExpungeAsync()
    {
      CheckSelected();

      return Client.StartTaskExclusively(() => ProcessResult(Client.Session.Expunge()),
                                         CancellationToken.None);
    }
#endif

    /*
     * MoveMessagesTo
     */
    public void MoveMessagesTo(ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapMailboxInfo destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(ImapMessageFetchAttributeOptions.None,
                                           uid,
                                           uids),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(searchCriteria,
                                           null,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               GetMessages(searchCriteria,
                                           encoding,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

#if ASYNC_OPS
    public Task MoveMessagesToAsync(ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           GetMessages(ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task MoveMessagesToAsync(ImapMailboxInfo destinationMailbox,
                                    long uid,
                                    params long[] uids)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           GetMessages(ImapMessageFetchAttributeOptions.None,
                                                       uid,
                                                       uids),
                                           destinationMailbox);
    }

    public Task MoveMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           GetMessages(searchCriteria,
                                                       null,
                                                       ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task MoveMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    Encoding encoding,
                                    ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           GetMessages(searchCriteria,
                                                       encoding,
                                                       ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }
#endif

    /*
     * CopyMessagesTo
     */
    public void CopyMessagesTo(ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapMailboxInfo destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(ImapMessageFetchAttributeOptions.None,
                                           uid,
                                           uids),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(searchCriteria,
                                           null,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailboxInfo destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               GetMessages(searchCriteria,
                                           encoding,
                                           ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

#if ASYNC_OPS
    public Task CopyMessagesToAsync(ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           GetMessages(ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task CopyMessagesToAsync(ImapMailboxInfo destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           GetMessages(ImapMessageFetchAttributeOptions.None,
                                                       uid,
                                                       uids),
                                           destinationMailbox);
    }

    public Task CopyMessagesToAsync(ImapSearchCriteria searchCriteria,
                               ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           GetMessages(searchCriteria,
                                                       null,
                                                       ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task CopyMessagesToAsync(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailboxInfo destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           GetMessages(searchCriteria,
                                                       encoding,
                                                       ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }
#endif

    private void MoveOrCopyMessagesToCore(bool copy,
                                          ImapMessageInfoList messages,
                                          ImapMailboxInfo destinationMailbox)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      if (copy)
        messages.CopyTo(destinationMailbox);
      else
        messages.MoveTo(destinationMailbox);
    }

#if ASYNC_OPS
    private Task MoveOrCopyMessagesToAsyncCore(bool copy,
                                               ImapMessageInfoList messages,
                                               ImapMailboxInfo destinationMailbox)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      if (copy)
        return messages.CopyToAsync(destinationMailbox);
      else
        return messages.MoveToAsync(destinationMailbox);
    }
#endif
  }
}
